/*-------------------------------------------------------------------------
	Copyright (C) 2008-2009 American Megatrends Inc.
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/jiffies.h>
#include <asm/io.h>
#include <asm/irq.h>
#include "i2c-data.h"
#include "i2c-hardware.h"
#ifdef ARBITRATION_SUPPORT
	#include "i2c-arbitration.h"
#else
	#define RESTART_SHORT_TIMER(val) do { } while(0)
#endif

#define ENABLE_STOP_DETECT

/* Enable this only for serious debugging */
//#define DEBUG

/* Prototypes */
irqreturn_t i2c_process_handler(int bus);

/* Disable further interrupt processing and wake the waiting process */
#define EXIT_INTERRUPT(b, status, s) {                               \
	if(s == STATE_IDLE) {                                            \
		i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET | \
		                           I2C_SERIAL_SCL_HIGH |             \
		                           I2C_SERIAL_SDA_HIGH),             \
		    (UINT32)i2c_vsc_io_base(b) + VSC_I2CX_BIT_CNTRL); }      \
	RESTART_SHORT_TIMER(RESTART_TIMER_VALUE);                        \
	vsc_data_ptr[b].op_status = status;                              \
	vsc_data_ptr[b].state = s;                                       \
	if (status) {                                                    \
		wake_up_interruptible(&vsc_data_ptr[b].vsc_wait); }          \
	return IRQ_HANDLED;                                              \
}


irqreturn_t i2c_handler(int this_irq, void *dev_id)
{
	volatile UINT32 ior;
	irqreturn_t vsc_irq = IRQ_NONE;

	ior = (i2c_vsc_read_reg((UINT32)INT_OBSERVE_REG));

	/* Check if the interrupt is generated by I2C0 */
	if (ior & I2C_OBSERVE_INT_0) {
		if(IRQ_HANDLED == i2c_process_handler(0)) {
			vsc_irq = IRQ_HANDLED;
		}
	}
	
	/* Check if the interrupt is generated by I2C1 */
	if (ior & I2C_OBSERVE_INT_1) {   
		if (IRQ_HANDLED == i2c_process_handler(1)) {
	    vsc_irq = IRQ_HANDLED;
		}
	}

	/* Check if the interrupt is generated by I2C2 */
	if (ior & I2C_OBSERVE_INT_2) {
		if (IRQ_HANDLED == i2c_process_handler(2)) {
			   vsc_irq = IRQ_HANDLED;
		}
	}
	
	/* Check if the interrupt is generated by I2C3 */
	if (ior & I2C_OBSERVE_INT_3) {
		if (IRQ_HANDLED == i2c_process_handler(3)) {
			vsc_irq = IRQ_HANDLED;
		}
	}
	
	/* Check if the interrupt is generated by I2C4 */
	if (ior & I2C_OBSERVE_INT_4) {
		if (IRQ_HANDLED == i2c_process_handler(4)) {
			vsc_irq = IRQ_HANDLED;
		}
	}
	
	/* Check if the interrupt is generated by I2C5 */
	if (ior & I2C_OBSERVE_INT_5) {
		if (IRQ_HANDLED == i2c_process_handler(5)) {
			vsc_irq = IRQ_HANDLED;
		}
	}
    
	return vsc_irq;
}

irqreturn_t i2c_process_handler(int bus)
{
	UINT32 volatile reg_read;
	UINT32 volatile status_reg_read;
	UINT32 volatile cntrl_bit_reg_read;
#ifdef ARBITRATION_SUPPORT
	u8 val_mler, val_mrer, val_mlbr;
#endif
	status_reg_read = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_STATUS);
	reg_read = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
	cntrl_bit_reg_read = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

	i2c_bus_status_ptr[bus].int_cnt ++;

#ifdef ARBITRATION_SUPPORT
	if (bus == I2C_BUS_2) {
		supfpga_mm_reg8_read(SUPFPGA_MLER_REG, &val_mler);
		supfpga_mm_reg8_read(SUPFPGA_MRER_REG, &val_mrer);
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);

		 /* if short/long timer expired or lost bus request/grant */
		if ((val_mler & (MLER_LOC_REL_SHORT_TMO | MLER_LOC_REL_LONG_TMO)) || (0 == (val_mlbr & (MLBR_LOC_BUS_REQ | MLBR_LOC_BUS_ACT))))  {
#ifdef DEBUG
			printk(KERN_INFO "jiffies=%lu - I2C%d: Local/Remote errors detected inside interrupt!!!\n", jiffies, bus);
#endif
			EXIT_INTERRUPT(bus, OP_STATUS_FPGA_TIMEOUT, STATE_IDLE);
		 } 
	}
#endif

#ifdef DEBUG
#ifdef ARBITRATION_SUPPORT
	if (bus == I2C_BUS_2) {
		printk(KERN_ERR "%ums - I2C%d:%02x:%02x:%02x:%02x:MLER=%04x:MRER=%04x:MLBR=%04x\n", jiffies_to_msecs(jiffies), bus, (u8)status_reg_read, (u8)reg_read, (u8)vsc_data_ptr[bus].slave_addr, (u8)vsc_data_ptr[bus].state, val_mler, val_mrer, val_mlbr);
	}
	else {
#endif
		printk(KERN_ERR "%ums - I2C%d:%02x:%02x:%02x:%02x\n", jiffies_to_msecs(jiffies), bus, (u8)status_reg_read, (u8)reg_read, (u8)vsc_data_ptr[bus].slave_addr, (u8)vsc_data_ptr[bus].state);
#ifdef ARBITRATION_SUPPORT
	}
#endif
#endif
	/* If stop condition occurred (we only enable stop condition detection just before WE issue a stop condition) */
	if (status_reg_read & (UINT32)I2C_STOP_DETECT_INT_STATUS) {

		/* Disable stop detect interrupt */
		i2c_disable_stop_detection(bus);
#ifdef ENABLE_STOP_DETECT
#ifdef DEBUG
		printk(KERN_INFO "%ums - I2C%d: detected stop\n", jiffies_to_msecs(jiffies), bus);
#endif
		/* If we were waiting for a stop condition (which we should be) */
		if (vsc_data_ptr[bus].state == STATE_WAIT_FOR_STOP) {
			/* return success back to calling thread notifying them that xmit is complete */		
			EXIT_INTERRUPT(bus, OP_STATUS_SUCCESS, STATE_IDLE);
		}
		else if (vsc_data_ptr[bus].state == STATE_WAIT_FOR_STOP_AFTER_NACK) {
			/* return NACK back to calling thread */		
			EXIT_INTERRUPT(bus, OP_STATUS_NACK, STATE_IDLE);
		}
		else 
#endif 
		{
#ifdef DEBUG
			printk(KERN_WARNING "%ums - I2C%d: Unexpected stop detected\n", jiffies_to_msecs(jiffies), bus);
#endif
			/* Since we are not waiting on stop or this stop is from another master (bus 4). Just return. */
			return IRQ_HANDLED;
		}
	}

	/* If interrupt pending */
	if (status_reg_read & (UINT32)I2C_SERIAL_INTERFACE_INT_STATUS) {   

		/* If bus error */
		if (reg_read & (UINT32)I2C_BUS_ERROR) {
#ifdef DEBUG
			printk(KERN_ERR "\nI2C%d: Bus error!\n", bus);
#endif
#if 1
			if ((vsc_data_ptr[bus].state == STATE_MASTER_XMIT) || 
				(vsc_data_ptr[bus].state == STATE_MASTER_XMIT_RECV) ||
				(vsc_data_ptr[bus].state == STATE_MASTER_RECV)) {

				/* Send stop condition */
				i2c_send_stop(bus);
			}
#endif
			/* Notify waiting thread (if one) of bus error */
			EXIT_INTERRUPT(bus, OP_STATUS_BUS_ERR, STATE_IDLE);

		/* If we are in process of writing data to a slave device (master transmit) */
		} else if ((vsc_data_ptr[bus].state == STATE_MASTER_XMIT) || 
				(vsc_data_ptr[bus].state == STATE_MASTER_XMIT_RECV)) {

			/* If this is the first byte to be transferred */
			if (vsc_data_ptr[bus].TX_index == 0) {

				/* If failed to generate start condition */
				if (reg_read & I2C_START_STATUS_NOT) {
#ifdef DEBUG
					printk(KERN_ERR "\nI2C%d: Start generation failed on write!\n", bus);
#endif
					/* Issue soft reset */
					i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET 
								| I2C_SERIAL_SCL_HIGH 
								| I2C_SERIAL_SDA_HIGH), 
								(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

					/* Notify waiting thread (if one) of error (we use arbitration lost here) */
					EXIT_INTERRUPT(bus, OP_STATUS_START_FAILED, STATE_IDLE);
				}
			}

			/* If arbitration lost */
			if (reg_read & I2C_LOST_ARBITRATION) {
#ifdef DEBUG
				printk(KERN_ERR "\nI2C%d: Arbitration lost on write\n", bus);
#endif
				/* Issue soft reset */
				i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET 
							| I2C_SERIAL_SCL_HIGH 
							| I2C_SERIAL_SDA_HIGH), 
							(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

				/* Notify waiting thread (if one) of arbitration lost */
				EXIT_INTERRUPT(bus, OP_STATUS_ARB_LOST, STATE_IDLE);
			}

			/* If NACK from slave */
			if (reg_read & I2C_LAST_RECEIVED_BIT) {
#ifdef DEBUG
				printk(KERN_ERR "I2C%d: Nack detected during write. Sending stop.\n", bus);
#endif               
#ifdef ENABLE_STOP_DETECT
				/* Enable stop detect interrupt */
				i2c_enable_stop_detection(bus);
#endif
				/* Send stop condition */
				i2c_send_stop(bus);
#ifdef ENABLE_STOP_DETECT
				/* Wait for completion of stop condition after NACK */
				EXIT_INTERRUPT(bus, OP_STATUS_WAIT, STATE_WAIT_FOR_STOP_AFTER_NACK);
#else
				/* Notify waiting thread (if one) of NACK */
				EXIT_INTERRUPT(bus, OP_STATUS_NACK, STATE_IDLE);
#endif				
			}

			/* If this is just a scan */
			if (vsc_data_ptr[bus].scan) {
#ifdef DEBUG
				printk(KERN_ERR "I2C%d: Device scan detected during write.  Sending stop.\n", bus); 
#endif
#ifdef ENABLE_STOP_DETECT
				/* Enable stop detect interrupt */
				i2c_enable_stop_detection(bus);
#endif
				/* Send stop condition */
				i2c_send_stop(bus);
#ifdef ENABLE_STOP_DETECT
				/* Wait for completion of stop condition */
				EXIT_INTERRUPT(bus, OP_STATUS_WAIT, STATE_WAIT_FOR_STOP);
#else
				/* Notify waiting thread (if one) that device was found */
				EXIT_INTERRUPT(bus, OP_STATUS_SUCCESS, STATE_IDLE);
#endif
			}

			/* If we still have data to send */
			if (vsc_data_ptr[bus].TX_index < vsc_data_ptr[bus].TX_len) {
#ifdef DEBUG
				printk(KERN_ERR "I2C%d: Wrote byte:%02x\n", bus, 
					vsc_data_ptr[bus].TX_data[vsc_data_ptr[bus].TX_index]);
#endif               
				/* Sending buffered data onto the I2C bus */
				i2c_vsc_write_reg((UINT32)(vsc_data_ptr[bus].TX_data[vsc_data_ptr[bus].TX_index]),
							(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_DATA);

				vsc_data_ptr[bus].TX_index++;
				
			} else {

				if (vsc_data_ptr[bus].state == STATE_MASTER_XMIT_RECV) {
#ifdef DEBUG
					printk(KERN_ERR "I2C%d: Repeated start for 0x%02x\n", bus, vsc_data_ptr[bus].slave_addr << 1);
#endif               
					/* Set state to master recv before issuing repeated start */
					vsc_data_ptr[bus].state = STATE_MASTER_RECV;
					vsc_data_ptr[bus].TX_len = 0;
					vsc_data_ptr[bus].MasterRX_index = 0;

					/* Perform repeated start condition */
					i2c_send_start(bus);

					/* Setup read operation slave address */
					i2c_vsc_write_reg((vsc_data_ptr[bus].slave_addr << 1) | READ_OPERATION_BIT, 
								(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_DATA);
				
				} else {
#ifdef DEBUG
					printk(KERN_ERR "I2C%d: Sending stop after write\n", bus);
#endif 
#ifdef ENABLE_STOP_DETECT
					/* Enable stop detect interrupt */
					i2c_enable_stop_detection(bus);
#endif
					/* Send stop condition */
					i2c_send_stop(bus);
#ifdef ENABLE_STOP_DETECT
					/* Wait for completion of stop condition */
					EXIT_INTERRUPT(bus, OP_STATUS_WAIT, STATE_WAIT_FOR_STOP);
#else
					/* Notify waiting thread (if one) transmission is complete */
					EXIT_INTERRUPT(bus, OP_STATUS_SUCCESS, STATE_IDLE);
#endif
				}
			}

		/* If we are in process of reading data from a slave device (master receive) */
		} else if (vsc_data_ptr[bus].state == STATE_MASTER_RECV) {

			/* If this is the first byte to be read */
			if (vsc_data_ptr[bus].MasterRX_index == 0) {

				/* If failed to generate start condition */
				if (reg_read & I2C_START_STATUS_NOT) {
#ifdef DEBUG
					printk(KERN_ERR "\nI2C%d: Start generation failed on read!\n", bus);
#endif
					/* Issue soft reset */
					i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET 
								| I2C_SERIAL_SCL_HIGH 
								| I2C_SERIAL_SDA_HIGH), 
								(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

					/* Notify waiting thread (if one) of error (we use arbitration lost here) */
					EXIT_INTERRUPT(bus, OP_STATUS_START_FAILED, STATE_IDLE);
				}
			}

			/* If arbitration lost */
			if (reg_read & I2C_LOST_ARBITRATION) {
#ifdef DEBUG
				printk(KERN_ERR "\nI2C%d: Arbitration lost on read\n", bus);
#endif
				/* Issue soft reset */
				i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET 
							| I2C_SERIAL_SCL_HIGH 
							| I2C_SERIAL_SDA_HIGH), 
							(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

				/* Notify waiting thread (if one) of arbitration lost */
				EXIT_INTERRUPT(bus, OP_STATUS_ARB_LOST, STATE_IDLE);
			}

			/* If this is the last byte to read */
			if (vsc_data_ptr[bus].MasterRX_index >= vsc_data_ptr[bus].MasterRX_len) {
#ifdef ENABLE_STOP_DETECT
				/* Enable stop detect interrupt */
				i2c_enable_stop_detection(bus);
#endif
				/* Send stop condition */
				i2c_send_stop(bus);

				if (vsc_data_ptr[bus].MasterRX_index == vsc_data_ptr[bus].MasterRX_len) {
					/* Read last byte */
					vsc_data_ptr[bus].MasterRX_data[vsc_data_ptr[bus].MasterRX_index - 1] = 
						i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus)+VSC_I2CX_DATA);
#ifdef DEBUG
					printk(KERN_ERR "I2C%d: Read last byte:%02x and send stop\n", bus, 
						vsc_data_ptr[bus].MasterRX_data[vsc_data_ptr[bus].MasterRX_index - 1]);
#endif
				} else {
					/* Received more than than we were expecting, stop now */
#ifdef DEBUG
					printk(KERN_ERR "I2C%d: Read extra byte:%02x ignore and send stop\n", bus, 
						vsc_data_ptr[bus].MasterRX_data[vsc_data_ptr[bus].MasterRX_index - 1]);
#endif
				}

#ifdef ENABLE_STOP_DETECT
				/* Wait for completion of stop condition */
				EXIT_INTERRUPT(bus, OP_STATUS_WAIT, STATE_WAIT_FOR_STOP);
#else
				/* Notify waiting thread (if one) transmission is complete */
				EXIT_INTERRUPT(bus, OP_STATUS_SUCCESS, STATE_IDLE);
#endif
			}

			/* If NACK from slave */
			if (reg_read & (UINT32)I2C_LAST_RECEIVED_BIT) {
#ifdef DEBUG
				printk(KERN_ERR "I2C%d: Nack detected during read. Sending stop.\n", bus);
#endif
#ifdef ENABLE_STOP_DETECT
				/* Enable stop detect interrupt */
				i2c_enable_stop_detection(bus);
#endif
				/* Send stop condition */
				i2c_send_stop(bus);
#ifdef ENABLE_STOP_DETECT
				/* Wait for completion of stop condition */
				EXIT_INTERRUPT(bus, OP_STATUS_WAIT, STATE_WAIT_FOR_STOP_AFTER_NACK);
#else
				/* Notify waiting thread (if one) of NACK */
				EXIT_INTERRUPT(bus, OP_STATUS_NACK, STATE_IDLE);
#endif
			}

			/* If we still have data to read */
			if (vsc_data_ptr[bus].MasterRX_index < vsc_data_ptr[bus].MasterRX_len) {

				/* Must do a dummy read for first byte */
				if (vsc_data_ptr[bus].MasterRX_index == 0) {
					/* Do a dummy read */
					u8 dummy_read;
					dummy_read = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus)+VSC_I2CX_DATA);
#ifdef DEBUG
					printk(KERN_ERR "I2C%d: Dummy read:%02x\n", bus, dummy_read);
#endif
				} else {

					/* Read byte from slave */
					vsc_data_ptr[bus].MasterRX_data[vsc_data_ptr[bus].MasterRX_index - 1] = 
						i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_DATA);
#ifdef DEBUG
					printk(KERN_ERR "I2C%d: Read byte:%02x\n", bus, 
						vsc_data_ptr[bus].MasterRX_data[vsc_data_ptr[bus].MasterRX_index - 1]);
#endif
				}
				/* If we are about to read last byte (during next interrupt)*/
				if (vsc_data_ptr[bus].MasterRX_len - vsc_data_ptr[bus].MasterRX_index == 1) {

					/* Clear acknowledge bit */
					i2c_vsc_write_reg((UINT32)(I2C_ENABLE_INTERFACE
								| I2C_ENABLE_INTERRUPT),
								(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
#ifdef DEBUG
					printk(KERN_ERR "I2C%d: Nack next byte we read\n", bus);
#endif
				}
				vsc_data_ptr[bus].MasterRX_index++;
		
			} else {
#ifdef DEBUG
				printk(KERN_ERR "I2C%d: Unexpected interrupt during master read\n", bus);
#endif
			}
		}  else {
			/* Usually, it doesn't fall in this area.
			 * If it falls in this area, we need to clear the pending interrupt and notify waiting thread.
			 * Issue soft reset to clear the pending interrupt*/
			i2c_vsc_write_reg((UINT32)(I2C_SERIAL_INTERFACE_SOFT_RESET
							| I2C_SERIAL_SCL_HIGH
							| I2C_SERIAL_SDA_HIGH),
							(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

			/* Notify waiting thread */
			EXIT_INTERRUPT(bus, OP_STATUS_UNKNOWN, STATE_IDLE);
		}

	/* Unkown interrupt condition */
	} else {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d: Unexpected interrupt!\n", bus);
#endif
	}

	return IRQ_HANDLED;
}


