/****************************************************************
****************************************************************
**                                                            **
**    (C)Copyright 2008-2009, American Megatrends Inc.        **
**                                                            **
**            All Rights Reserved.                            **
**                                                            **
**       5555 Oakbrook Parkway,Building 200, Norcross,        **
**                                                            **
**        Georgia - 30093, USA. Phone-(770)-246-8600.         **
**                                                            **
****************************************************************
**  mdcio driver implementation
**
**
**************************************************************/

/* Version Number of this module */
#define SAMPLE_MAJOR	1
#define SAMPLE_MINOR	0

#include <linux/module.h>
#include <linux/kernel.h>	/* printk() 		*/
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/ide.h>
#include <linux/delay.h>
#include <linux/irqflags.h>
#include "gpio_vsc.h"
#include "gpio_info.h"
#include "mdcio.h"

#define	dbgprint	if(Debug_Enable) printk

#define MDCIO_MAX_DEVICES	1
#define MDCIO_MAX_CHANNEL       (MDCIO_MAX_DEVICES - 1)

#define MDCIO_MAJOR	250
#define MDCIO_DEV_NAME	"mdcio"


#define CLK_LOW		0
#define CLK_HIGH	1

#define DATA_LOW	0
#define DATA_HIGH	1	 

/* Constants */
#define OPT_CTRL_REG	(*(unsigned long *)0xB02008A4)

/* Mdcio enable mask  */
#define MDCIO_ENABLE	0x00000C00

/* gpio pins for data and clock for channel 0 and 1*/
int mdcio_data_pin [ 2 ] = { 16 };
int mdcio_clk_pin [ 2 ] = { 17 }; 


/* Prototypes */
static int mdcio_open(struct inode * inode, struct file * filp);
static int drive_clock(int channel, int data);
static int data_out(int channel, int data);
static int data_in(int channel);
static int dir_out(int channel);
//static int dir_in(int channel);
static int open_drain(int channel);
static long mdcio_ioctl( struct file *file,
                     unsigned int cmd, unsigned long arg);
static int read_reg ( unsigned int channel,unsigned int phyaddr,unsigned int regnum );
static int write_reg ( unsigned int channel,unsigned int phyaddr, unsigned int regnum, unsigned int data );



/* Device structure */
static struct file_operations mdcio_fops = {
	owner:		THIS_MODULE,
	read:		NULL,
	write:		NULL,
	unlocked_ioctl:		mdcio_ioctl,
	open:		mdcio_open,
	release:	NULL,
};


/* semaphore to provide mutex */
struct semaphore g_mdcio_sem;

gpio_dev_t *gpio_dev = NULL;
extern void * get_gpio_dev(void);

/* Module Init Function*/
int init_mdcio_module(void)
{
	int 	err =0;
	Gpio_data_t	gpio_data;
	int i = 0;

	printk(KERN_ALERT "MDCIO  Driver Version %d.%d\n",SAMPLE_MAJOR,SAMPLE_MINOR);
	printk(KERN_ALERT "Copyright (c) 2007 American Megatrends Inc.\n");

	/* Init Mutex */
	init_MUTEX(&g_mdcio_sem);
	/*  Enable MDIO 0 and 1   */
/*	OPT_CTRL_REG |= MDCIO_ENABLE;*/
	OPT_CTRL_REG &= (~MDCIO_ENABLE);	 
	gpio_dev = get_gpio_dev();

	/* Initialize clock as output and data in open drain */	
	for ( i = 0; i < MDCIO_MAX_DEVICES; i++)
	{
		/* configure clock as output and set high */
		gpio_data.PinNum = mdcio_clk_pin[i];
		gpio_data.data = 1;
		vsc_setdir( gpio_dev, &gpio_data);
		drive_clock(i,CLK_LOW);
		
		/* configure data pin as open drain */
		open_drain(i);
	}
	
	if ( NULL == gpio_dev )
	{
		printk(KERN_ERR "MDCIO: Initializing gpio dev failed.\n");
		return -ENODEV;
	}

	/* ----  Register the character device ------------------- */
	if ((err = register_chrdev(MDCIO_MAJOR, MDCIO_DEV_NAME, &mdcio_fops)) < 0)
	{
		printk (KERN_ERR "MDCIO: failed to register device (err %d)\n", err);
		return -ENODEV;
	}
	return 0;
	
}


/* Module Exit Function*/
void exit_mdcio_module(void)
{
        printk(KERN_ALERT "Unloading  MDCIO Module ..............\n");
	
	/* Disable the mdcio interface */
/*	OPT_CTRL_REG &= (~MDCIO_ENABLE); */
        
	/* unregister the device */
        unregister_chrdev(MDCIO_MAJOR, MDCIO_DEV_NAME);

        printk(KERN_ALERT "MDCIO module unloaded sucessfully.\n");	
}

/* Open */
static int mdcio_open(struct inode * inode, struct file * filp)
{

        if(filp->f_flags & 010000000)
        {
                return 0;
        }
        filp->f_op =&mdcio_fops;

	return  0;
}



/* ioctl */
static long  mdcio_ioctl( struct file *file,
                     unsigned int cmd, unsigned long arg)

{
	mdcio_ioctl_t   pkt;

        if ( copy_from_user((void*)&pkt,(void*)arg,sizeof(mdcio_ioctl_t)))
        return -EFAULT;

	if (pkt.channelnum > MDCIO_MAX_CHANNEL) {
		printk(KERN_ERR "mdcio_ioctl: request received for invalid "
				"channel %d\n", pkt.channelnum);
		return -EINVAL;
	}

	down(&g_mdcio_sem);
         
	switch ( cmd )
	{
		case READ_REG:
			pkt.data = read_reg ( pkt.channelnum,pkt.devaddr, pkt.regnum);			
			if (copy_to_user((void*)arg,(void*) &pkt , sizeof(mdcio_ioctl_t)))
			{
				up(&g_mdcio_sem);
			 	return -EFAULT;
			}
			break;
		
		case WRITE_REG:
			write_reg(pkt.channelnum,pkt.devaddr, pkt.regnum, pkt.data);
                      	break;

		
		default:
			break;
	}	
	up(&g_mdcio_sem);
	return 0;

}

/* Read register from switch */
static int read_reg (unsigned int channel,unsigned int phyaddr ,unsigned int regnum )
{
	/* 0x60 in the MSB indicates start condition with read operation  */
	unsigned long data = 0x6000;
	unsigned int reg_data = 0;
	int temp = 0;
	int i =0;

	/* embed the phyaddr and regnum with start condition and read cmd */
	phyaddr = phyaddr << 7;
	data = data | phyaddr;
	regnum = regnum << 2;
	data = data | regnum;


    /* stop any preemption, this allow us to run uninterrupted since we're timing sensitive now */
    preempt_disable();  

    /*Sending Initial 32 bit preamble*/
	open_drain(channel);
	for ( i =0; i < 32; i++)
        {
		drive_clock(channel, CLK_LOW);
		drive_clock(channel, CLK_HIGH);
        }
	
	/* set data pin direction out */
	dir_out(channel);

	/* or start condition, phyaddr, regnum, ta and receive data */
	for ( i =0; i < 14; i++)
        {
		drive_clock(channel, CLK_LOW);
                if ( data & (1 << (15 -i) ))
                        data_out(channel, DATA_HIGH);
                else
                        data_out(channel, DATA_LOW);

                drive_clock(channel, CLK_HIGH);
        }
	
	/* do the turn around */
	open_drain(channel);
	
	drive_clock(channel, CLK_LOW);
        drive_clock(channel, CLK_HIGH);

	temp = 2;	
	i = 0;
	while ( i < 25 )
	{
		temp = data_in(channel);
	        if (temp==0)
        	{
			break;
		}
		i++;
	}	

	if ( i  >= 25 )
	{
		printk(KERN_ERR "Device not responding\n");
        preempt_enable();
        return -1;
	}
	drive_clock(channel, CLK_LOW);

	/* Read the data */
	for ( i=0; i < 16; i++ )
	{
		drive_clock(channel,CLK_HIGH);
		temp = data_in(channel);
		temp = (temp << ( 15 - i));
		reg_data = reg_data | temp;
		drive_clock(channel, CLK_LOW );
	}

	/*Sending Last 32 Idle bits */
	open_drain(channel);
	for ( i =0; i < 32; i++)
        {
		drive_clock(channel, CLK_HIGH);
		drive_clock(channel, CLK_LOW);
        }
    
    preempt_enable();

	return reg_data;
}


/* Write in the switch register */
static int write_reg (unsigned int channel,unsigned int phyaddr ,unsigned int regnum,unsigned int reg_data )
{
	/* 50 in the msb indicates start condition with write cmd */
	unsigned long data = 0x50000000;
	int i = 0;

	/* embed the phyaddr and regnum with the start condition and write cmd*/
        phyaddr = phyaddr << 23;
        data = data | phyaddr;
        regnum = regnum << 18;
        data = data | regnum;
        
	/*Sending Initial 32 bit preamble*/
	open_drain(channel);
	for ( i =0; i < 32; i++)
        {
		drive_clock(channel, CLK_LOW);
		drive_clock(channel, CLK_HIGH);
        }

	/* set data pin direction out */
        dir_out(channel);
	
	/* or start condition,phy addr, reg num, data*/
	for ( i =0; i < 14; i++)
	{
		drive_clock(channel, CLK_LOW);
		if ( data & (1 << (31 -i) ))	
		{
			data_out(channel, DATA_HIGH);
		}
		else
		{
			data_out(channel, DATA_LOW);
		}
		drive_clock(channel, CLK_HIGH);		
	}
	
	/*do the turn around*/
	drive_clock(channel,CLK_LOW);
	data_out(channel,DATA_HIGH);
	drive_clock(channel, CLK_HIGH);
	
	drive_clock(channel, CLK_LOW);
	data_out(channel, DATA_LOW);
	drive_clock(channel, CLK_HIGH);

	/* write the data  */	
	for ( i=0; i < 16; i++ )
	{
		drive_clock( channel, CLK_LOW);
		if ( reg_data & ( 1 << (15 -i)))
			data_out(channel, DATA_HIGH);
		else
			data_out(channel, DATA_LOW);
		drive_clock( channel, CLK_HIGH);
	}

	/* put the data line in opendrain */
	open_drain(channel);
	for ( i =0; i < 32; i++)
        {
		drive_clock(channel, CLK_LOW);
		drive_clock(channel, CLK_HIGH);
        }

	return 0;
}

/* Drive Clock  */
static int drive_clock(int channel,int data)
{
	Gpio_data_t gpio_data;
	
	gpio_data.PinNum = mdcio_clk_pin [ channel ];
	gpio_data.data = data;

	vsc_writegpio( gpio_dev, &gpio_data);
	udelay(50);

	return 0;
}


/* Direction Out */
static int dir_out(int channel)
{
        Gpio_data_t gpio_data;

        gpio_data.PinNum = mdcio_data_pin [ channel ];
        gpio_data.data = 1;

        vsc_setdir( gpio_dev, &gpio_data);

	udelay(50);
        return 0;
}
#if 0
// Commented out as it is unused in this file
/* Drection In */
static int dir_in(int channel)
{
        Gpio_data_t gpio_data;

        gpio_data.PinNum = mdcio_data_pin [ channel ];
        gpio_data.data = 0;

        vsc_setdir( gpio_dev, &gpio_data);

        return 0;
}
#endif

/* Data Out */
static int data_out(int channel, int data )
{
	Gpio_data_t gpio_data;
	
	gpio_data.PinNum = mdcio_data_pin [ channel ];
	gpio_data.data = data;

	vsc_writegpio( gpio_dev, &gpio_data);

	return 0;
}


/* Data In */
static int data_in(int channel)
{
        Gpio_data_t gpio_data;

        gpio_data.PinNum = mdcio_data_pin [ channel ];
        gpio_data.data = 0;

        vsc_readgpio( gpio_dev, &gpio_data);

        return gpio_data.data;
}
 

/* Open drain */
static int open_drain(int channel)
{
	Gpio_data_t gpio_data;

	gpio_data.PinNum =  mdcio_data_pin[ channel ];
	gpio_data.data = 0;	
	
	vsc_setdir( gpio_dev, &gpio_data);

	return 0;
}

 int 
mdcio_read_reg ( unsigned int channel, unsigned int phyaddr, unsigned int regnum, unsigned int *data)
{
	int retval; 

	if (channel > MDCIO_MAX_CHANNEL) {
		printk(KERN_ERR "mdcio_read_reg: request received for invalid "
				"channel %d\n", channel);
		/* Cant return an error code here, so return -1 */
		return -1;
	}

	*data = read_reg(channel, phyaddr, regnum);

	if ( *data == -1 )
		retval  = -1;
	else 
		retval = 0;

	return retval;
}

int 
mdcio_write_reg ( unsigned int channel, unsigned int phyaddr, unsigned int regnum, unsigned int data )
{
	if (channel > MDCIO_MAX_CHANNEL) {
		printk(KERN_ERR "mdcio_read_reg: request received for invalid "
				"channel %d\n", channel);
		/* Cant return an error code here, so return -1 */
		return -1;
	}

	return write_reg(channel, phyaddr, regnum, data);
}

/*--------------------------- Module Information Follows --------------------------*/
module_init (init_mdcio_module);
module_exit (exit_mdcio_module);

EXPORT_SYMBOL_GPL(mdcio_read_reg);
EXPORT_SYMBOL_GPL(mdcio_write_reg);

MODULE_AUTHOR("American Megatrends Inc");
MODULE_DESCRIPTION("This is MCIO driver module.");
MODULE_LICENSE("GPL");
