--- linux-2.6.15.4-mips/arch/mips/vsc452/int.c	1969-12-31 19:00:00.000000000 -0500
+++ linux-2.6.15.4-mips-vsc452/arch/mips/vsc452/int.c	2008-08-25 10:39:11.984739244 -0400
@@ -0,0 +1,120 @@
+/*
+ * Carsten Langgaard, carstenl@mips.com
+ * Copyright (C) 2000, 2001, 2004 MIPS Technologies, Inc.
+ * Copyright (C) 2001 Ralf Baechle
+ *
+ *  This program is free software; you can distribute it and/or modify it
+ *  under the terms of the GNU General Public License (Version 2) as
+ *  published by the Free Software Foundation.
+ *
+ *  This program is distributed in the hope it will be useful, but WITHOUT
+ *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
+ *  for more details.
+ *
+ *  You should have received a copy of the GNU General Public License along
+ *  with this program; if not, write to the Free Software Foundation, Inc.,
+ *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
+ *
+ */
+#include <linux/init.h>
+#include <linux/kernel.h>
+#include <asm/system.h>
+#include <asm/irq_cpu.h>
+#include <linux/interrupt.h>
+#include <asm/mach-vsc452/vsc452.h>
+
+void __init arch_init_irq(void)
+{
+//        if (!cpu_has_veic)
+                mips_cpu_irq_init();
+}
+
+static inline int clz(unsigned long x)
+{
+        __asm__(
+        "       .set    push                                    \n"
+        "       .set    mips32                                  \n"
+        "       clz     %0, %1                                  \n"
+        "       .set    pop                                     \n"
+        : "=r" (x)
+        : "r" (x));
+
+        return x;
+}
+
+/*
+ *  * Version of ffs that only looks at bits 12..15.
+ *   */
+static inline unsigned int irq_ffs(unsigned int pending)
+{
+#if defined(CONFIG_CPU_MIPS32) || defined(CONFIG_CPU_MIPS64)
+        return -clz(pending) + 31 - CAUSEB_IP;
+#else
+        unsigned int a0 = 7;
+        unsigned int t0;
+
+        t0 = pending & 0xf000;
+        t0 = t0 < 1;
+        t0 = t0 << 2;
+        a0 = a0 - t0;
+        pending = pending << t0;
+
+        t0 = pending & 0xc000;
+        t0 = t0 < 1;
+        t0 = t0 << 1;
+        a0 = a0 - t0;
+        pending = pending << t0;
+
+        t0 = pending & 0x8000;
+        t0 = t0 < 1;
+        /* t0 = t0 << 2; */
+        a0 = a0 - t0;
+        /* pending = pending << t0; */
+
+        return a0;
+#endif
+}
+
+
+/*
+ ** IRQs on the Malta board look basically (barring software IRQs which we
+ ** don't use at all and all external interrupt sources are combined together
+ ** on hardware interrupt 0 (MIPS IRQ 2)) like:
+ **
+ **      MIPS IRQ        Source
+ **      --------        ------
+ **             0        Software (ignored)
+ **             1        Software (ignored)
+ **             2        Combined hardware interrupt (hw0)
+ **             3        Hardware (ignored)
+ **             4        Hardware (ignored)
+ **             5        Hardware (ignored)
+ **             6        Hardware (ignored)
+ **             7        R4k timer (what we use)
+ **
+ ** We handle the IRQ according to _our_ priority which is:
+ **
+ ** Highest ----     R4k Timer
+ ** Lowest  ----     Combined hardware interrupt
+ **
+ ** then we just return, if multiple IRQs are pending then we will just take
+ ** another exception, big deal.
+ **/
+
+asmlinkage void plat_irq_dispatch(void)
+{
+        unsigned int pending;
+        int irq;
+        pending = read_c0_cause() & read_c0_status() & ST0_IM;
+
+        irq = irq_ffs(pending);
+
+        if (irq >= 0)
+	{
+                do_IRQ(MIPS_CPU_IRQ_BASE + irq);
+	}
+        else
+                spurious_interrupt();
+}
+
