--- uboot/common/cmd_systemoff.c	2014-12-01 15:42:06.727628000 -0500
+++ uboot/common/cmd_systemoff.c	2014-12-01 15:45:46.993196000 -0500
@@ -0,0 +1,173 @@
+/*  Copyright 2000, 2001
+* Wolfgang Denk, DENX Software Engineering, wd@denx.de.
+*
+* See file CREDITS for list of people who contributed to this
+* project.
+*
+* This program is free software; you can redistribute it and/or
+* modify it under the terms of the GNU General Public License as
+* published by the Free Software Foundation; either version 2 of
+* the License, or (at your option) any later version.
+*
+* This program is distributed in the hope that it will be useful,
+* but WITHOUT ANY WARRANTY; without even the implied warranty of
+* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+* GNU General Public License for more details.
+*
+* You should have received a copy of the GNU General Public License
+* along with this program; if not, write to the Free Software
+
+* Foundation, Inc., 59 Temple Place, Suite 330, Boston,
+* MA 02111-1307 USA
+*
+*/
+
+/*
+* This command holds the system in reset by asserting bit 12
+* of the ALGE register in the support FPGA.
+* This command was added to allow the NAND flash to be 
+* programmed for the first time without being interrupted
+* by the BIOS SP recovery mechanism.
+*
+*/
+
+#include <common.h>
+#include <config.h>
+#include <command.h>
+#include <asm/io.h>
+#include <i2c.h>
+#include <vsc452/vsc452_i2c.h>
+
+#define GPIO_PORT0              0x10200a00
+#define GPIO_OFFSET             4
+#define OUTPUT_LOW              0x00000000
+#define OUTPUT_HIGH             0x00000001
+#define GPIO(port, number)      GPIO_PORT0 + (port *0x40) + (number * GPIO_OFFSET)
+#define SP_CRIT_SIG_EN          GPIO(0,14)
+#define PS_GOOD			GPIO(1,3)
+
+#define write32(addr, val)      writel(val, 0xa0000000 + (addr))
+#define read32(addr)            readl(0xa0000000 + (addr))
+
+#define SEQCPLD_MUX_SLAVE_ADDR     0x77
+#define SEQCPLD_ALGE_SLAVE_ADDR    0x17
+#define SEQCPLD_ALSR_SLAVE_ADDR    0x18
+#define SEQCPLD_I2C_BUS            0x03
+#define SEQCPLD_MUX_CHANNEL        0x02
+#define SEQCPLD_POWER_DOWN_MASK    0x7F
+#define SEQCPLD_POWER_DOWN_DEASSERT_MASK 0x80
+#define SEG_CPLD_ALGE_BYTE         0x00
+#define CRIT_SIG_STROBE_DELAY      1000  // 1 ms
+#define POWER_GOOD		   0x01
+
+#define AGENT_LITE_SIZE		   0x05
+
+void strobe_seq_cpld_crit_sig (void);
+
+int do_systemoff( cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
+{
+	unsigned char buffer[AGENT_LITE_SIZE] = {0,};
+	unsigned char offset = 0;
+	unsigned char len = 0;
+	unsigned char alen = 0;
+	int ret_val = 0;
+
+	/* Init the I2C driver to use bus 3 */
+	if((ret_val = vsc452_i2c_init(SEQCPLD_I2C_BUS)) < 0) {
+		printf("I2C init failed\n");
+	}
+
+	if (ret_val >= 0) {
+		/* Set up the Sequence CPLD MUX */
+		buffer[0] = SEQCPLD_MUX_CHANNEL;
+		len = 1;
+		if ((ret_val = i2c_write(SEQCPLD_MUX_SLAVE_ADDR,offset,alen,buffer,len)) < 0) {
+			printf("Failed setting up Support FPGA I2C Mux\n");
+		}
+	}
+
+	if (ret_val >= 0) {
+		/* Read the ALGE register */
+		if ((ret_val = i2c_read(SEQCPLD_ALGE_SLAVE_ADDR,offset,alen,buffer,len)) < 0) {
+			printf("Failed to read the ALGE register\n");
+		}
+		else {
+			buffer[SEG_CPLD_ALGE_BYTE] = buffer[SEG_CPLD_ALGE_BYTE] & SEQCPLD_POWER_DOWN_MASK;
+		}
+	}
+	/* Assert the Power Off bit in the Seq CPLD */
+	if (ret_val >= 0) {
+		/* Write to the ALGE register */
+		if ((ret_val = i2c_write(SEQCPLD_ALGE_SLAVE_ADDR,offset,alen,buffer,len)) < 0) {
+			printf("Failed to write to the ALGE register\n");
+		}
+	}
+
+	if (ret_val >= 0) {
+		/* Read back the ALGE register */
+		if ((ret_val = i2c_read(SEQCPLD_ALGE_SLAVE_ADDR,offset,alen,buffer,len)) < 0) {
+			printf("Failed to read back the ALGE register\n");
+		}
+		else {
+			if ((buffer[SEG_CPLD_ALGE_BYTE] & ~SEQCPLD_POWER_DOWN_MASK) == 0) {
+				/* Strobe the critical signal GPIO */
+				strobe_seq_cpld_crit_sig();
+			}
+			else {
+				printf("Write to ALGE register did not take %02x\n",buffer[SEG_CPLD_ALGE_BYTE]);
+			}
+		}
+	}
+
+	udelay(160000);
+
+	/* Deassert the Power Off bit in the Seq CPLD */
+	if (ret_val >= 0) {
+                /* Read the ALGE register */
+                if ((ret_val = i2c_read(SEQCPLD_ALGE_SLAVE_ADDR,offset,alen,buffer,len)) < 0) {
+                        printf("Failed to read the ALGE register on deassert\n");
+                }
+                else {
+                        buffer[SEG_CPLD_ALGE_BYTE] = buffer[SEG_CPLD_ALGE_BYTE] | SEQCPLD_POWER_DOWN_DEASSERT_MASK;
+                }
+        }
+
+	if (ret_val >= 0) {
+                /* Write to the ALGE register */
+                if ((ret_val = i2c_write(SEQCPLD_ALGE_SLAVE_ADDR,offset,alen,buffer,len)) < 0) {
+                        printf("Failed to write to the ALGE register on deassert\n");
+                }
+        }
+
+        if (ret_val >= 0) {
+                /* Read back the ALGE register */
+                if ((ret_val = i2c_read(SEQCPLD_ALGE_SLAVE_ADDR,offset,alen,buffer,len)) < 0) {
+                        printf("Failed to read back the ALGE register on deassert\n");
+                }
+                else {
+                        if ((buffer[SEG_CPLD_ALGE_BYTE] & SEQCPLD_POWER_DOWN_DEASSERT_MASK) == SEQCPLD_POWER_DOWN_DEASSERT_MASK) {
+                                /* Strobe the critical signal GPIO */
+                                strobe_seq_cpld_crit_sig();
+				printf("System is powered down!\n");
+                        }
+                        else {
+                                printf("Write to ALGE register did not take %02x\n",buffer[SEG_CPLD_ALGE_BYTE]);
+                        }
+                }
+        }
+
+	return ret_val;
+}
+
+void strobe_seq_cpld_crit_sig (void)
+{
+	write32(SP_CRIT_SIG_EN,OUTPUT_LOW);
+        udelay(CRIT_SIG_STROBE_DELAY);
+        write32(SP_CRIT_SIG_EN,OUTPUT_HIGH);
+}
+
+ U_BOOT_CMD(
+        systemoff, 1,      1,      do_systemoff,
+        "systemoff - Powers down the CPU complex\n",
+        "Powers down the CPU complex to allow safe NAND flash programming\n"
+);
